(*  Press 
    Copyright (C) 2005, 2006 M. Sighireanu and L. Allali

    This file is part of Press.

    Press is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Press is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Press; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*)

(*
 *  Implementation of the reachability algorithm modulo term equality 
 *  for Process Rewrite Systems
 *
 *  File:         $RCSfile: prs_reach_eq.ml,v $
 *  Author:       M. Sighireanu
 *  Version:      $Revision: 1.4 $
 *  Date:         $Date: 2006/05/31 19:44:34 $
 *)


(* ******************************************************************
 * Import of modules.
 * ******************************************************************)

open Symbol
open Prs_reach_util


(* ******************************************************************
 * Local functions.
 * ******************************************************************)

(** Build transitions given by the algorithm rule 1.
 *  Algorithm rule 1: q -> qT for all states q in qSet
 *)
let build_reach_1 a qSet =
	let sigma = (Automaton.get_alphabet a) and
	    sigmap = (Automaton.get_state_ops a) in
	let rule1 trs q = 
		(let term_q = (state2term sigma sigmap q)
		 and term_qT = (to_term_T sigma sigmap q)
		in
		let trs1 = (Automaton.normalize_epsilon term_q term_qT trs) in
		begin
			(* print_message ("trs1 = "^(PRS.to_string trs1)) *)
			(PRS.union trs trs1) (* TODO: speed up *)
		end
		)
	in
	let trs1 = (List.fold_left rule1 
			(Automaton.get_transitions a)
			((States.to_list qSet)))
	in
	let a1 = (Automaton.make_automaton 
			(Automaton.get_alphabet a)
			(Automaton.get_state_ops a)
			(Automaton.get_states a)
			(Automaton.get_final_states a)
			(PRS.empty) 
			trs1)
	in 
	begin
	(print_debug_aut a1 "rule1" true);
	a1
	end
;;

(** Build transitions given by the algorithm rule 2.
 *  Algorithm rule 2: if 0 ->* q in d0 then 0 -> qnil in d'.
 *)
let build_reach_2 a d0 q0 = 
	let t0 = (string2term (Automaton.get_alphabet a) 
			      (Automaton.get_state_ops a) idleOp) in
	let a0 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        d0) in
	let rule2 q = (Prs_reach_util.if_run_then_build_trs 
			a0 t0 q 
			a [t0] [(to_state_nil q)]) in
	let trs2 = (List.fold_left PRS.union_fast PRS.empty (List.map rule2 (States.to_list q0))) in
	let a2 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs2)) (* TODO: speed up *)
	in
	begin
	(print_debug_aut a2 "rule2" true);
	a2
	end
;;

(** Build transitions given by the algorithm rule 3.
 *  Algorithm rule 3:
 *  if t1 -> t2 in r and exists q in Q0 U QR s.t. t1 ->* qT in a then
 *  	add q(t2)T -> qT, q(t2)nil -> qnil and q(t2)T -> qnil iff t1 = 0
 *)
let build_reach_3 a q0 qR r = 
        let sigma = (Automaton.get_alphabet a) and
            sigmap = (Automaton.get_state_ops a) in
	(** action to take for each rule in r and each q in q0 U qR *)
	let rule3_r_s rule q = (
		let t1 = (PRS.lhs rule) and
                    t2 = (PRS.rhs rule) and
		    t0 = (string2term sigma sigmap idleOp) in
		let qT = (to_state_T q) 
		and qnil = (to_state_nil q)
		and qt2T = (string2term sigma sigmap ((to_state_name t2)^suffixT))
		and qt2nil = (string2term sigma sigmap ((to_state_name t2)^suffixNil))
                in begin
	  	   let srcL0 = [qt2T; qt2nil] and dstL0 = [qT; qnil] in
		   let srcL,destL = (
			if Terms.equal t1 t0 then 
				(srcL0@[qt2T], dstL0@[qnil])
			else
				(srcL0, dstL0)
			) in
		    (Prs_reach_util.if_run_then_build_trs 
			a t1 qT
			a srcL destL)  	
		end
		) in 
	(* action to take for each rule in r *)
	let rule3_r q0 qR rule = (
		List.fold_left PRS.union_fast
				PRS.empty 
				(List.map (rule3_r_s rule) ((States.to_list q0)@(States.to_list qR)))
		) in
	let trs3 = (
		List.fold_left PRS.union 
				PRS.empty 
				(List.map (rule3_r q0 qR) (PRS.to_list r))) in
	let a3 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs3)) (* TODO: speed up *)
	in
	begin
	(print_debug_aut a3 "rule3" true);
	a3
	end
;;

(** Build transitions given by the algorithm rule 3.
 *  Algorithm rule 4: seq(q1,q2) -> q in d0 U dR for all q1,q2,q in Q0 U QR. 
 *)
let build_reach_4 a d0 dR = 
        let sigma = (Automaton.get_alphabet a) and
            sigmap = (Automaton.get_state_ops a) in
	let rule4 r = (
		let lhs = (PRS.lhs r) and 
                    rhs = (PRS.rhs r) in
		(if ((Symbol.to_string (Terms.top_symbol lhs)) = seqOp) &&
		    (Terms.is_special rhs) then
			let q1 = (List.hd (Terms.list_arguments lhs)) and
			    q2 = (List.hd (List.tl (Terms.list_arguments lhs)))
			in
			let str_rule4a1 = (seqOp^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^") -> "^(Terms.to_string rhs)^suffixT^" ") and
			    str_rule4a2= (seqOp^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixT^") -> "^(Terms.to_string rhs)^suffixT^" ") and
			    str_rule4a3= (seqOp^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixT^" ") and
			    str_rule4b = (seqOp^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixNil^" ") and
			    str_rule4c = (seqOp^"("^(Terms.to_string q1)^suffixT^", "^(Terms.to_string q2)^") -> "^(Terms.to_string rhs)^suffixT^" ") in
			let rule4a1 = (string2trs sigma sigmap str_rule4a1) and
			    rule4a2 = (string2trs sigma sigmap str_rule4a2) and
			    rule4a3 = (string2trs sigma sigmap str_rule4a3) and
			    rule4b = (string2trs sigma sigmap str_rule4b) and
			    rule4c = (string2trs sigma sigmap str_rule4c) 
			in
			(List.fold_left PRS.union PRS.empty [rule4a1; rule4a2; rule4a3; rule4b; rule4c])
		 else
			(PRS.empty)
		)
		) in
	let trs4 = (List.fold_left PRS.union PRS.empty (List.map rule4 ((PRS.to_list d0)@(PRS.to_list dR)))) in
	let a4 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs4)) (* TODO: speed up *)
	in
	begin
	(print_debug_aut a4 "rule4" true);
	a4
	end
;;

(** Build transitions given by the algorithm rule 5.
 *  Algorithm rule 5: par(q1,q2) -> q in d0 U dR for all q1,q2,q in Q0 U QR.
 *)
let build_reach_5 a d0 dR = 
        let sigma = (Automaton.get_alphabet a) and
            sigmap = (Automaton.get_state_ops a) in
	let rule5 r = (
		let lhs = (PRS.lhs r) and 
                    rhs = (PRS.rhs r) in
                let op = (Symbol.to_string (Terms.top_symbol lhs)) in
		(if ((op = parOp) || (op = par0Op)) &&
		    (Terms.is_special rhs) then
			let q1 = (List.hd (Terms.list_arguments lhs)) and
			    q2 = (List.hd (List.tl (Terms.list_arguments lhs)))
			in
			let str_rule5a = (op^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixT^" ") and
			    str_rule5b= (
				op^"("^(Terms.to_string q1)^", "^(Terms.to_string q2)^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^", "^(Terms.to_string q2)^suffixT^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixNil^", "^(Terms.to_string q2)^suffixT^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixT^", "^(Terms.to_string q2)^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixT^", "^(Terms.to_string q2)^suffixNil^") -> "^(Terms.to_string rhs)^suffixT^" "^
			    	op^"("^(Terms.to_string q1)^suffixT^", "^(Terms.to_string q2)^suffixT^") -> "^(Terms.to_string rhs)^suffixT^" ") in
			let rule5a = (string2trs sigma sigmap str_rule5a) and
			    rule5b = (string2trs sigma sigmap str_rule5b) 
			in
			(PRS.union rule5a rule5b)
		 else
			(PRS.empty)
		)
		) in
	let trs5 = (List.fold_left PRS.union PRS.empty (List.map rule5 ((PRS.to_list d0)@(PRS.to_list dR)))) in
	let a5 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs5)) (* TODO: speed up *)
	in
	begin
	(print_debug_aut a5 "rule5" true);
	a5
	end
;;

(** Algorithm rules 1 to 5: one pass *)
let build_reach_1_5 a0 d0 dR q0 qR r =
	let a1 = (build_reach_1 a0 (States.union q0 qR)) in
	let a2 = (build_reach_2 a1 d0 q0) in
	let a3 = (build_reach_3 a2 q0 qR r) in
	let a4 = (build_reach_4 a3 d0 dR) in
	let a5 = (build_reach_5 a4 d0 dR) in
	(** Note: rule 6 is not allowed by Timbuk tree automata *)
		a5
;;

(** Main algorithm: fixed point computation. *)
let rec
    build_reach a d0 dR q0 qR r = 
	begin
	(print_new_iteration ());
	no_iteration := !no_iteration + 1;
	let newa = (build_reach_1_5 a d0 dR q0 qR r) in
	(** Note: rule 6 is not allowed by Timbuk tree automata *)
	if (List.length (PRS.to_list (Automaton.get_transitions a))) = 
	   (List.length (PRS.to_list (Automaton.get_transitions newa))) then
		a
	else
		(build_reach newa d0 dR q0 qR r)
	end
;;

(* ******************************************************************
 * Main algorithm.
 * ******************************************************************)

let build_aut_reach_eq a0 r =
	begin
	(print_message "build reachability automaton modulo term equality");
	(let sigma = (Automaton.get_alphabet a0) in
	 let q0 = (Automaton.get_states a0) in
	 let delta0 = (Automaton.get_transitions a0) in
	 let aR = (build_aut_prs sigma r) in
	 let qR = (Automaton.get_states aR) in
	 let deltaR = (Automaton.get_transitions aR) in
	 let str_sigmap = (build_states_reach "Q'" (States.union q0 qR)) in
	 let sigmap = (string2stateOps str_sigmap) in
	 let qp = (string2stateSet sigma sigmap str_sigmap) in
	 let str_sigmafp = (build_states_reach "F'" (Automaton.get_final_states a0)) in
	 let sigmafp = (string2stateOps str_sigmafp) in
	 let fp = (string2stateSet sigma sigmafp str_sigmafp) in
	 let reachp = (Automaton.make_automaton sigma sigmap qp fp (PRS.empty) (PRS.union delta0 deltaR)) in
	 begin
         (print_debug_aut reachp "init" false);
	 let lfp = (build_reach reachp delta0 deltaR q0 qR r) and
	     lfp_name = (!filename_debug)^"_clean.aut.tbk" in
	 begin
	 (print_debug_aut lfp "full" false);
         (print_message "clean the automaton");
         let lfp_red = (Automaton.clean lfp) in
         begin
		(if (!option_equiv = 0) then
			(print_message ("=====> see file "^lfp_name))
		 else ()
		)
		;
	 	(print_debug_aut lfp_red "clean" false);
		lfp_red
	 end
	 end
	 end
	)
	end
;;

