(*  Press 
    Copyright (C) 2005, 2006 M. Sighireanu and L. Allali

    This file is part of Press.

    Press is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Press is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Press; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*)

(*
 *  Implementation of the reachability algorithm modulo ~s.
 *
 *  File:         $RCSfile: prs_reach_s.ml,v $
 *  Author:       M. Sighireanu
 *  Version:      $Revision: 1.4 $
 *  Date:         $Date: 2006/05/31 19:44:34 $
 *)


(* ******************************************************************
 * Import of modules.
 * ******************************************************************)

open Symbol
open Prs_reach_util
open Prs_reach_0


(* ******************************************************************
 * Local functions.
 * ******************************************************************)

(** Definitions for manipulation of seqctx states.
 *)

(** Constant used.  *)
let str_seqctx = "seqctx"
;;

(** Build the name of state from q and X => seqctx_q_x_.
 *)
let build_state_seqctx q var =
	(str_seqctx^"_"^(Terms.to_string q)^"_"^var^"_")
;;

(** Build the states corresponding to seqctx (q,X).  *)
let build_states_seqctx sigma qSet =
	let build_states_seqctx_q_var q var = (
		if (Prs_reach_util.is_user_var var) then
		begin
			(* (print_message var); *)
                	(build_state_seqctx q var)
		end
		else
			""
		)
	in
	let build_states_seqctx_q sigma q =
		(String.concat " " (List.map (build_states_seqctx_q_var q) (Sigma.to_string_list sigma)))
	in
	let str_states_seqctx = (String.concat " " (List.map (build_states_seqctx_q sigma) (States.to_list qSet)))
	in
		str_states_seqctx
;;


(** Definitions used by the algorithm.
 *)
(** Algorithm rule alpha1: q -> qT forall q in qNull.
 *)
let build_reach_s_1 a qNull = (
	Prs_reach_eq.build_reach_1 a qNull)
;;

(** Algorithm rule alpha2: 
 *  if Y ->*(a) q and seq(X,Y) -> t in R then qtT -> (q,X).
 *)
let rec
    build_reach_s_2 a r qNull =
	let trs2 = (List.fold_left PRS.union PRS.empty
			(List.map (build_reach_s_2_r a qNull) (PRS.to_list r)))
	in
	let a2 = (Automaton.make_automaton 
			(Automaton.get_alphabet a)
			(Automaton.get_state_ops a)
       	         	(Automaton.get_states a)
       	         	(Automaton.get_final_states a)
                	(PRS.empty)
			(PRS.union (Automaton.get_transitions a) trs2))
	in
        begin
                (print_debug_aut a2 "rule2" true);
                a2
        end

and build_reach_s_2_r a qNull r =
    let lhs = (PRS.lhs r) and
        rhs = (PRS.rhs r) in
        (if ((Symbol.to_string (Terms.top_symbol lhs)) = seqOp) then
	    (* assert: PRS in normal form *)
	    let varX = (List.hd (Terms.list_arguments lhs)) and
		varY = (List.hd (List.tl (Terms.list_arguments lhs))) in
	    let rhsT = (string2term 
	                 (Automaton.get_alphabet a)
       		         (Automaton.get_state_ops a)
       		         ((to_state_name rhs)^suffixT)) in
		List.fold_left PRS.union PRS.empty
			(List.map (build_reach_s_2_r_q a varY rhsT varX) (States.to_list qNull)) 
	 else
		PRS.empty
	)

and build_reach_s_2_r_q a tY tT tX q =
	let qX = (string2state (build_state_seqctx q (Terms.to_string tX)))
	in
	(Prs_reach_util.if_run_then_build_trs a tY q a [tT] [qX])
;;


(** Algorithm rule alpha3:
 *  if exists t1->t2 in r and q in qNull s.t.:
 *  a) t1->*(a) qT then qt2T -> qT (@see Prs_reach_0.build_reach_0_rule2)
 *  b) t1->*(a) (q,X) then qt2T -> (q,X)
 *)
let rec 
    build_reach_s_3 a r qNullT qPair =
    let trs3 = List.fold_left PRS.union PRS.empty
                (List.map (build_reach_3_r a (States.union_disjoint qNullT qPair)) (PRS.to_list r))
    in
    let a3 = (Automaton.make_automaton
		(Automaton.get_alphabet a)
		(Automaton.get_state_ops a)
		(Automaton.get_states a)
		(Automaton.get_final_states a)
		(PRS.empty)
		(PRS.union (Automaton.get_transitions a) trs3))
    in
  	begin
        (print_debug_aut a3 "rule3" true);
        a3
  	end

and build_reach_3_r_q aRun srcRun srcBuild dst =
       	 (Prs_reach_util.if_run_then_build_trs aRun srcRun dst aRun [srcBuild] [dst])

and build_reach_3_r aRun qSet rule =
	 let lhsTerm = (PRS.lhs rule) and
             rhsTerm = (string2term (Automaton.get_alphabet aRun)
                             	    (Automaton.get_state_ops aRun)
                                    ((to_state_name (PRS.rhs rule)^suffixT)))
 	 in
         (List.fold_left PRS.union PRS.empty
		(List.map (build_reach_3_r_q aRun lhsTerm rhsTerm) (States.to_list qSet)))
;;


(** Algorithm rules alpha4 + alpha5:
 *  forall op(q1, q2) -> q in deltaNull (for 4 op = seqOp, for 5 op = parOp)
 *  4a) seq(q1T, q2) -> qT 		  5a) par(q1T,q2T) -> qT
 *      @see Prs_reach_0.build_reach_0_rule3_4
 *  4b) forall X seq(q1X, q2) -> qX
 *  4c) forall X if 0 ->*(a) q1X then q2T -> qX
 *  4d) if 0 ->*(a) q1T then q2T -> qT	  5b) same
 *					  5c) if 0 ->*(a) q2T then q1T -> qT
 *	@see Prs_reach_0.build_reach_0_rule6
 *  4e) forall X . if X ->*(a) q1T then q2X -> qT
 *  4f) forall X, X' . if X ->*(a) q1X' then q2X -> qX'
 *)
let rec
    build_reach_s_4_5 a qNullT qPair deltaNull = 
	let trs4 = List.fold_left PRS.union PRS.empty 
			(List.map (build_reach_s_4_5_r a qNullT qPair)
				  (PRS.to_list deltaNull))
	in
	let a4 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs4))
  	in
  	begin
        (print_debug_aut a4 "rule45" true);
        a4
	end

and build_reach_s_4_5_r a qNull qPair rule =
    let sigma = (Automaton.get_alphabet a) and
        sigmap = (Automaton.get_state_ops a) in
    let lhs = (PRS.lhs rule) and
        rhs = (PRS.rhs rule) in
    let op = Symbol.to_string (Terms.top_symbol lhs) and
	t0 = (string2term sigma sigmap idleOp)
    in

    (if (((op = seqOp) || (op = parOp) || (op = par0Op)) &&
         (Terms.is_special rhs))
    then
	let q1 = (List.hd (Terms.list_arguments lhs)) and
            q2 = (List.hd (List.tl (Terms.list_arguments lhs))) and
             q = (Terms.get_special rhs) 
	in
        let str_q1 = (Terms.to_string q1) and
            str_q2 = (Terms.to_string q2) and
            str_q  = (Terms.to_string q) 
	in
        let str_q1T = str_q1^suffixT and
            str_q2T = str_q2^suffixT and
            str_qT  = str_q^suffixT 
	in
	let q1T = (string2term sigma sigmap str_q1) and
	    q2T = (string2term sigma sigmap str_q2) and
	     qT = (string2term sigma sigmap str_q) 
	in
	(if (op = seqOp) then
	let trs_4a = (build_reach_s_4a sigma sigmap op str_q1 str_q2 str_q)
	and trs_4d = (build_reach_s_4d a op t0 str_q1T q2T str_qT)
	and trs_4bcef = (List.fold_left PRS.union PRS.empty
			(List.map 
			(build_reach_s_4bcef_X a sigma sigmap op t0 q1 q1T q2 q2T q qT)
			(List.filter Prs_reach_util.is_user_var 
					(Sigma.to_string_list sigma))
			)
        		)
	in 
	(List.fold_left PRS.union PRS.empty
			[trs_4a; trs_4d; trs_4bcef])
	else (* op = parOp or par0Op *)
	let trs_5a = (build_reach_s_5a sigma sigmap op str_q1 str_q2 str_q)
	and trs_5bc = (build_reach_s_5bc a op t0 str_q1T str_q2T q2T q1T str_qT)
	in 
	(PRS.union trs_5a trs_5bc)
	)
     else
	PRS.empty
     )

(* 4a. seq(q1T, q2) -> qT
 *	@see Prs_reach_0.build_reach_0_rule3_4
 *)
and build_reach_s_4a sigma sigmaq op str_q1 str_q2 str_q =
	if (op = seqOp) then
       	  (string2trs sigma sigmaq (seqOp^"("^str_q1^suffixT^","^str_q2^") -> "^str_q^suffixT))
	else
	  PRS.empty

(*  4b. forall X seq(q1X, q2) -> qX
 *)
and build_reach_s_4b_X sigma sigmap str_q1X str_q2 str_qX =
	 (string2trs sigma sigmap 
			(seqOp^"("^str_q1X^","^str_q2^") -> "^str_qX))

(*  4c. forall X if 0 ->*(a) q1X then q2T -> qX
 *)
and build_reach_s_4c_X a t0 str_q1X q2T str_qX = 
	(Prs_reach_util.if_run_then_build_trs 
			a t0 (string2state str_q1X)
			a [q2T] [(string2state str_qX)])

(*  4d. if 0 ->*(a) q1T then q2T -> qT
 *	@see Prs_reach_0.build_reach_0_rule6_r
 *)
and build_reach_s_4d a op t0 str_q1T q2T str_qT = 
	if (op = seqOp) then
	(Prs_reach_util.if_run_then_build_trs 
				a t0 (string2state str_q1T) 
				a [q2T] [(string2state str_qT)]
        ) 
	else
	  PRS.empty

(*  4e. forall X . if X ->*(a) q1T then q2X -> qT
 *)
and build_reach_s_4e_X a sigma sigmap tX str_q1T str_q2X str_qT = 
	(Prs_reach_util.if_run_then_build_trs
			a tX (string2state str_q1T)
			a [(string2term sigma sigmap str_q2X)] [(string2state str_qT)])

(*  4f. forall X, X' . if X ->*(a) q1X' then q2X -> qX'
 *)
and build_reach_s_4f_X_Xp a sigma sigmap tX q1 str_q2X q vXp = 
	(let q1Xp = (string2state (build_state_seqctx q1 vXp)) and
	     qXp = (string2state (build_state_seqctx q vXp)) 
	 in
	 (Prs_reach_util.if_run_then_build_trs 
			a tX q1Xp
			a [(string2term sigma sigmap str_q2X)] [qXp])
	)  

and build_reach_s_4bcef_X a sigma sigmap op t0 q1 q1T q2 q2T q qT vX =
	(* assert op = seqOp *)
	let str_q1X = (build_state_seqctx q1 vX) and
            str_q2X = (build_state_seqctx q2 vX) and
            str_qX  = (build_state_seqctx q vX) and
	    tX = (string2term sigma sigmap vX)
	in
	let trs_4b = (build_reach_s_4b_X sigma sigmap str_q1X (Terms.to_string q2) str_qX)
	and trs_4c = (build_reach_s_4c_X a t0 str_q1X q2T str_qX)
	and trs_4e = (build_reach_s_4e_X a sigma sigmap tX (Terms.to_string q1T) str_q2X (Terms.to_string qT))
	and trs_4f = (List.fold_left PRS.union PRS.empty
			(List.map 
				(build_reach_s_4f_X_Xp a sigma sigmap tX q1 str_q2X q)
				(List.filter Prs_reach_util.is_user_var 
					(Sigma.to_string_list sigma))
			)
		      )
	in
	(List.fold_left PRS.union PRS.empty [trs_4b;trs_4c;trs_4e;trs_4f])

(* 5a. par(q1T,q2T) -> qT 
 *	@see Prs_reach_0.build_reach_0_rule3_4
 *)
and build_reach_s_5a sigma sigmaq op str_q1 str_q2 str_q =
	if (op = parOp) || (op = par0Op) then
	  (string2trs sigma sigmaq (op^"("^str_q1^suffixT^","^str_q2^suffixT^") -> "^str_q^suffixT))
        else PRS.empty

(*  5b. if 0 ->*(a) q1T then q2T -> qT
 *  5c. if 0 ->*(a) q2T then q1T -> qT 
 *)
and build_reach_s_5bc a op t0 str_q1T str_q2T q2T q1T str_qT =
	if (op = parOp) || (op = par0Op) then
          (PRS.union
         	(Prs_reach_util.if_run_then_build_trs 
					a t0 (string2state str_q1T) 
					a [q2T] [(string2state str_qT)])
                (Prs_reach_util.if_run_then_build_trs 
					a t0 (string2state str_q2T)
					a [q1T] [(string2state str_qT)])
	  )
        else PRS.empty
;;


(** Algorithm rule alpha6:
 *)
let build_reach_s_6 a qNull deltaNull = 
	(* since deltaNull is normalized, 
         * there are no rules q -> q' in deltaNull *)
	a
;;

(** Algorithm rule alpha7:
 *  seq(qNullT, qT) -> qT forall q in qSet.
 *)
let build_reach_s_7 a qNull = 
	let trs7 = (Prs_reach_0.build_reach_0_rule7 a qNull) in
        let a7 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        trs7)
        in
        begin
        (print_debug_aut a7 "rule7" true);
        a7
        end
;;

(** Build reach automaton by applying rules alpha1-7.
 *  @param a    initially is automaton with (Q~, Sigma, F~, deltaNull).
 *  @param r    prs
 *)
let rec
    build_reach_s a r qNull qNullT qPair deltaNull = 
        (* rules alpha1, alpha6, and alpha7 do not need iteration *)
	let a1 = build_reach_s_1 a qNull in 
        let a16 = build_reach_s_6 a1 qNull deltaNull in 
        let a167 = build_reach_s_7 a16 qNull in 
	build_reach_s_2345 a167 r qNull qNullT qPair deltaNull

and build_reach_s_2345 a r qNull qNullT qPair deltaNull =
        begin
        (print_new_iteration ());
	no_iteration := !no_iteration + 1;
	(let a2 = (build_reach_s_2 a r qNull) in
	 let a3 = (build_reach_s_3 a2 r qNullT qPair) in 
	 let a45 = (build_reach_s_4_5 a3 qNullT qPair deltaNull) in 
         if (List.length (PRS.to_list (Automaton.get_transitions a))) =
            (List.length (PRS.to_list (Automaton.get_transitions a45))) then
                a
         else
                (build_reach_s_2345 a45 r qNull qNullT qPair deltaNull)
	)
        end
;;


(* ******************************************************************
 * Main algorithm.
 * ******************************************************************)

let build_aut_reach_s a0 r = 
	begin
        (print_message "build reachability automaton modulo idle neutrality \n\t\t and seq associativity");
        (let aNull = (build_closure_0 a0 r) in
	 let sigma = (Automaton.get_alphabet aNull) and
	     sigmaqNull = (Automaton.get_state_ops aNull) and
	     qNull = (Automaton.get_states aNull) and
	     qFNull = (Automaton.get_final_states aNull) and
	     deltaNull = (Automaton.get_transitions aNull) in
	 let str_sigmaqNullT = (build_states_reachT qNull) in
	 let str_qFNullT = (build_states_reachT qFNull) in
         let sigmaqNullT = (string2stateOps str_sigmaqNullT) in
	 let str_sigmaqPair = (build_states_seqctx sigma qNull) in
	 let sigmaqPair = (string2stateOps str_sigmaqPair) in
         let sigmaq = (Sigma.union_fast sigmaqNull 
			(Sigma.union_fast sigmaqNullT sigmaqPair)) in
	 let qNullT = (string2stateSet sigma sigmaq str_sigmaqNullT) in
	 let qPair = (string2stateSet sigma sigmaq str_sigmaqPair) in
	 let qT = (States.union_disjoint qNull (States.union_disjoint qNullT qPair)) in
	 let qFT = (States.union_disjoint qFNull (string2stateSet sigma sigmaq str_qFNullT)) in
	 let reachs = (Automaton.make_automaton 
			sigma 
			sigmaq
			qT 
			qFT 
			(PRS.empty) 
			deltaNull) in
	 begin
	 (print_debug_aut reachs "init" false);
	 let lfp = (build_reach_s reachs r qNull qNullT qPair deltaNull) and
	     lfp_name = (!filename_debug)^"_clean.aut.tbk" in
	 begin
	 (print_debug_aut lfp "full" false);
	 (print_message "clean the automaton");
	 let lfp_red = (Automaton.clean lfp) in
         begin
                (if (!option_equiv = 2) then
                        (print_message ("=====> see file "^lfp_name))
                 else ()
                )
		;
	 	(print_debug_aut lfp_red "clean" false);
                lfp_red
	 end
         end
	 end
	)
	end
;;

(* End of $RCSfile: prs_reach_s.ml,v $ *)
